import unittest

import numpy as np

from sigllm.primitives.postprocessing import aggregate_rolling_window


class FormatAsStringTest(unittest.TestCase):
    def test_aggregate_rolling_window_shape_1_1_5(self):
        data = np.array([[[1, 2, 3, 4, 5]]])
        expected = np.array([1, 2, 3, 4, 5])

        output = aggregate_rolling_window(data)

        np.testing.assert_array_equal(output, expected)

    def test_aggregate_rolling_window_shape_1_3_5(self):
        data = np.array([[[1, 2, 3, 4, 5], [2, 3, 4, 5, 6], [3, 4, 5, 6, 7]]])
        expected = np.array([2, 3, 4, 5, 6])

        output = aggregate_rolling_window(data)

        np.testing.assert_array_equal(output, expected)

    def test_aggregate_rolling_window_shape_1_5_1(self):
        data = np.array([[[1], [2], [3], [4], [5]]])
        expected = np.array([3])

        output = aggregate_rolling_window(data)

        np.testing.assert_array_equal(output, expected)

    def test_aggregate_rolling_window_shape_3_1_5(self):
        data = np.array([[[1, 2, 3, 4, 5]], [[2, 3, 4, 5, 6]], [[3, 4, 5, 6, 7]]])
        expected = np.array([1, 2, 3, 4, 5, 6, 7])

        output = aggregate_rolling_window(data)

        np.testing.assert_array_equal(output, expected)

    def test_aggregate_rolling_window_shape_3_2_5(self):
        data = np.array([
            [[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]],
            [[2, 3, 4, 5, 6], [3, 4, 5, 6, 7]],
            [[3, 4, 5, 6, 7], [4, 5, 6, 7, 8]],
        ])
        expected = np.array([1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5])

        output = aggregate_rolling_window(data)

        np.testing.assert_array_equal(output, expected)



# -*- coding: utf-8 -*-

"""Prompter post-processing module.

This module contains functions that help filter LLMs results to get the final anomalies.
"""

import ast
import re

import numpy as np

PATTERN = r'\[([\d\s,]+)\]'


def _clean_response(text):
    text = text.strip().lower()
    text = re.sub(r',+', ',', text)

    if 'no anomalies' in text or 'no anomaly' in text:
        return ''

    return text


def _parse_list_response(text):
    clean = _clean_response(text)

    # match anything that consists of digits and commas
    match = re.search(PATTERN, clean)

    if match:
        values = match.group(1)
        values = [val.strip() for val in values.split(',') if val.strip()]
        return ','.join(values)

    return ''


def _parse_interval_response(text):
    clean = _clean_response(text)
    match = re.finditer(PATTERN, clean)

    if match:
        values = list()
        for m in match:
            interval = ast.literal_eval(m.group())
            if len(interval) == 2:
                start, end = ast.literal_eval(m.group())
                values.extend(list(range(start, end + 1)))

        return values

    return []


def parse_anomaly_response(X, interval=False):
    """Parse a list of lists of LLM responses to extract anomaly values and format them as strings.

    Args:
        X (List[List[str]]):
            List of lists of response texts from the LLM in the format
            "Answer: no anomalies" or "Answer: [val1, val2, ..., valN]."
            values must be within brackets.
        interval (bool):
            Whether to parse the response as a list "Answer: [val1, val2, ..., valN]."
            or list of intervals "Answer: [[s1, e1], [s2, e2], ..., [sn, en]]."

    Returns:
        List[List[str]]:
            List of lists of parsed responses where each element is either
            "val1,val2,...,valN" if anomalies are found, or empty string if
            no anomalies are present.
    """
    method = _parse_list_response
    if interval:
        method = _parse_interval_response

    result = []
    for response_list in X:
        parsed_list = [method(response) for response in response_list]
        result.append(parsed_list)

    return result


def val2idx(y, X):
    """Convert detected anomalies values into indices.

    Convert windows of detected anomalies values into an array of all indices
    in the input sequence that have those values.

    Args:
        y (ndarray):
            A 3d array containing detected anomalous values from different
            responses of each window.
        X (ndarray):
            rolling window sequences.

    Returns:
        List([ndarray]):
            A 3d array containing detected anomalous indices from different
            responses of each window.
    """
    idx_list = []
    for anomalies_list, seq in zip(y, X):
        idx_win_list = []
        for anomalies in anomalies_list:
            mask = np.isin(seq, anomalies)
            indices = np.where(mask)[0]
            idx_win_list.append(indices)
        idx_list.append(idx_win_list)
    idx_list = np.array(idx_list, dtype=object)

    return idx_list


def find_anomalies_in_windows(y, alpha=0.5):
    """Get the final list of anomalous indices of each window.

    Choose anomalous index in the sequence based on multiple LLM responses

    Args:
        y (ndarray):
            A 3d array containing detected anomalous values from different
            responses of each window.
        alpha (float):
            Percent of votes needed for an index to be anomalous. Default to `0.5`.

    Returns:
        ndarray:
            A 2-dimensional array containing final anomalous indices of each windows.
    """
    if isinstance(y, list):
        y = np.array(y, dtype=object)

    idx_list = []
    for samples in y:
        min_vote = np.ceil(alpha * len(samples))

        flattened_res = np.concatenate(samples.tolist())

        unique_elements, counts = np.unique(flattened_res, return_counts=True)

        final_list = unique_elements[counts >= min_vote]

        idx_list.append(final_list)
    idx_list = np.array(idx_list, dtype=object)

    return idx_list


def merge_anomalous_sequences(y, first_index, window_size, step_size, beta=0.5):
    """Merge sequences.

    Get the final list of anomalous indices of a sequence when merging all rolling windows.

    Args:
        y (ndarray):
            A 2-dimensional array containing anomalous indices of each window.
        first_index (ndarray):
            A 1-dimensional array contaning the first index of each window.
        window_size (int):
            Length of each window
        step_size (int):
            Indicating the number of steps the window moves forward each round.
        beta (float):
            Percent of windows needed for index to be anomalous. Default to `0.5`.

    Return:
        ndarray:
            A 1-dimensional array containing final anomalous indices.
    """
    anomalies = [arr + first_idx for (arr, first_idx) in zip(y, first_index)]

    min_vote = np.ceil(beta * window_size / step_size)

    flattened_res = np.concatenate(anomalies)

    unique_elements, counts = np.unique(flattened_res, return_counts=True)

    final_list = unique_elements[counts >= min_vote]

    return np.sort(final_list)


def format_anomalies(y, timestamp, padding_size=50):
    """Format anomalies into intervals.

    Convert list of anomalous indices to list of intervals
    by padding to both sides and merge overlapping.

    Args:
        y (ndarray):
            A 1-dimensional array of indices. Can be empty if no anomalies are found.
        timestamp (ndarray):
            List of full timestamp of the signal.
        padding_size (int):
            Number of steps to pad on both sides of a timestamp point. Default to `50`.

    Returns:
        List[Tuple]:
            List of intervals (start, end, score). Empty list if no anomalies are found.
    """
    # Handle empty array case
    if len(y) == 0:
        return []

    y = y[y < len(timestamp)].astype(int)
    y = timestamp[y]  # Convert list of indices into list of timestamps
    start, end = timestamp[0], timestamp[-1]
    interval = timestamp[1] - timestamp[0]
    intervals = []
    for timestamp in y:
        intervals.append((
            max(start, timestamp - padding_size * interval),
            min(end, timestamp + padding_size * interval),
        ))
    if not intervals:
        return []

    intervals.sort(key=lambda x: x[0])  # Sort intervals based on start time
    merged_intervals = [intervals[0]]  # Initialize merged intervals with the first interval

    for current_interval in intervals[1:]:
        previous_interval = merged_intervals[-1]

        # If the current interval overlaps with the previous one, merge them
        if current_interval[0] <= previous_interval[1]:
            previous_interval = (
                previous_interval[0],
                max(previous_interval[1], current_interval[1]),
            )
            merged_intervals[-1] = previous_interval
        else:
            merged_intervals.append(current_interval)  # Append the current interval if no overlap

    merged_intervals = [(interval[0], interval[1], 0) for interval in merged_intervals]

    return merged_intervals



import json
import os
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock, call, mock_open, patch

import pandas as pd
from orion.evaluation import CONTEXTUAL_METRICS as METRICS
from orion.evaluation import contextual_confusion_matrix
from pytest import fixture

from sigllm.benchmark import (
    _augment_hyperparameters,
    _evaluate_signal,
    _get_pipeline_directory,
    _get_pipeline_hyperparameter,
    _run_job,
    benchmark,
)


@fixture
def base_hyperparameters():
    return {
        'time_segments_aggregate#1': {'interval': 1},
        'rolling_window#1': {'window_size': 100},
        'other_param': 'other_value',
    }


def test__augment_hyperparameters_without_few_shot(base_hyperparameters):
    result = _augment_hyperparameters(base_hyperparameters, few_shot=False)

    assert result == base_hyperparameters


def test__augment_hyperparameters_with_few_shot(base_hyperparameters):
    result = _augment_hyperparameters(base_hyperparameters, few_shot=True)

    assert len(result) == 4
    assert result['time_segments_aggregate#1'] == base_hyperparameters['time_segments_aggregate#1']
    assert result['time_segments_aggregate#2'] == base_hyperparameters['time_segments_aggregate#1']
    assert result['rolling_window#1'] == base_hyperparameters['rolling_window#1']
    assert result['other_param'] == base_hyperparameters['other_param']


def test__augment_hyperparameters_with_empty_hyperparameters(base_hyperparameters):
    result = _augment_hyperparameters({}, few_shot=True)

    assert result == {}


def test__augment_hyperparameters_with_multiple_aggregate_params():
    hyperparameters = {
        'time_segments_aggregate#1': 'value1',
        'time_segments_aggregate#3': 'value3',
        'other_param': 'other_value',
    }

    result = _augment_hyperparameters(hyperparameters, few_shot=True)

    assert result['time_segments_aggregate#1'] == 'value1'
    assert result['time_segments_aggregate#2'] == 'value1'
    assert result['time_segments_aggregate#3'] == 'value3'
    assert result['time_segments_aggregate#4'] == 'value3'
    assert result['other_param'] == 'other_value'


@patch('os.path.isfile')
def test__get_pipeline_directory(mock_isfile):
    pipeline_path = '/fake/path/pipeline_dir/pipeline.json'
    mock_isfile.return_value = True

    expected = '/fake/path/pipeline_dir'
    result = _get_pipeline_directory(pipeline_path)

    assert result == expected


@patch('sigllm.benchmark.get_pipelines_paths')
@patch('os.path.isfile')
def test__get_pipeline_directory_search_pipeline_paths(mock_isfile, mock_pipeline_paths):
    pipeline_name = 'pipeline'
    base_path = '/fake/path/pipeline/'
    mock_pipeline_paths.return_value = [base_path]

    # first isfile check should be False to enter the pipeline paths loop
    # second isfile check should be True when finding the json file
    mock_isfile.side_effect = [False, True]

    expected = '/fake/path/pipeline'
    result = _get_pipeline_directory(pipeline_name)

    assert result == expected

    mock_isfile.assert_has_calls([
        call(pipeline_name),
        call(os.path.join(base_path, 'pipeline.json')),
    ])
    mock_pipeline_paths.assert_called_once()


class TestGetPipelineHyperparameter(TestCase):
    @classmethod
    def setup_class(cls):
        cls.pipeline_name = 'test_pipeline'
        cls.dataset_name = 'test_dataset'

        cls.temp_dir = tempfile.mkdtemp()

        cls.base_hyperparameters = {'default_param': 'default_value'}

        cls.dataset_hyperparameters = {'test_dataset': {'dataset_param': 'dataset_value'}}

        cls.pipeline_hyperparameters = {
            'test_dataset': {'test_pipeline': {'pipeline_param': 'pipeline_value'}}
        }

    def test__get_pipeline_hyperparameter_exist_direct(self):
        """Test when hyperparameters are directly provided"""
        hyperparameters = {'param1': 'value1', 'param2': 'value2'}

        result = _get_pipeline_hyperparameter(
            hyperparameters, self.dataset_name, self.pipeline_name
        )

        self.assertEqual(result, hyperparameters)

    def test__get_pipeline_hyperparameter_exist_nested(self):
        """Test when hyperparameters are nested by dataset and pipeline"""
        hyperparameters = {
            'test_dataset': {'test_pipeline': {'param1': 'value1', 'param2': 'value2'}}
        }

        result = _get_pipeline_hyperparameter(
            hyperparameters, self.dataset_name, self.pipeline_name
        )

        expected = {'param1': 'value1', 'param2': 'value2'}
        self.assertEqual(result, expected)

    def test__get_pipeline_hyperparameter_do_not_exist(self):
        """Test when no hyperparameters are provided"""
        result = _get_pipeline_hyperparameter(None, self.dataset_name, self.pipeline_name)

        self.assertIsNone(result)

    def test__get_pipeline_hyperparameter_dataset_dependent_hyperparameters(self):
        """Test when hyperparameters are dataset-specific"""
        hyperparameters = {
            'test_dataset': {'param1': 'dataset_value1', 'param2': 'dataset_value2'},
            'other_dataset': {'param1': 'other_value1', 'param2': 'other_value2'},
        }

        result = _get_pipeline_hyperparameter(
            hyperparameters, self.dataset_name, self.pipeline_name
        )

        expected = {'param1': 'dataset_value1', 'param2': 'dataset_value2'}
        self.assertEqual(result, expected)

        result = _get_pipeline_hyperparameter(
            hyperparameters, 'nonexistent_dataset', self.pipeline_name
        )

        self.assertEqual(result, hyperparameters)

    @patch('os.path.exists')
    @patch('json.load')
    def test__get_pipeline_hyperparameter_from_json_file(self, mock_json_load, mock_exists):
        """Test loading hyperparameters from a JSON file"""
        mock_exists.return_value = True
        mock_json_data = {'param1': 'file_value1', 'param2': 'file_value2'}
        mock_json_load.return_value = mock_json_data

        file_path = os.path.join(self.temp_dir, 'test_pipeline_test_dataset.json')

        with patch('builtins.open', mock_open(read_data=json.dumps(mock_json_data))):
            result = _get_pipeline_hyperparameter(file_path, self.dataset_name, self.pipeline_name)

        self.assertEqual(result, mock_json_data)

    @patch('sigllm.benchmark._get_pipeline_directory')
    @patch('os.path.exists')
    @patch('json.load')
    def test__get_pipeline_hyperparameter_none_with_pipeline_path(
        self, mock_json_load, mock_exists, mock_get_pipeline_dir
    ):
        """Test when hyperparameters is None but dataset and pipeline are specified.
        Should look for JSON file in pipeline directory."""
        pipeline_dir = '/fake/path/pipeline_dir'
        mock_get_pipeline_dir.return_value = pipeline_dir
        mock_exists.return_value = True

        expected_hyperparams = {'param1': 'auto_value1', 'param2': 'auto_value2'}
        mock_json_load.return_value = expected_hyperparams

        expected_file_path = os.path.join(
            pipeline_dir, f'{os.path.basename(pipeline_dir)}_{self.dataset_name.lower()}.json'
        )

        with patch('builtins.open', mock_open(read_data=json.dumps(expected_hyperparams))):
            result = _get_pipeline_hyperparameter(None, self.dataset_name, self.pipeline_name)

        mock_get_pipeline_dir.assert_called_once_with(self.pipeline_name)
        mock_exists.assert_called_once_with(expected_file_path)

        self.assertEqual(result, expected_hyperparams)

        mock_exists.return_value = False
        result = _get_pipeline_hyperparameter(None, self.dataset_name, self.pipeline_name)
        self.assertIsNone(result)

    def test__get_pipeline_hyperparameter_fallback_behavior(self):
        """Test the fallback behavior when dataset/pipeline specific params don't exist"""
        hyperparameters = {
            'default': {'param': 'default_value'},
            'test_dataset': {'param': 'dataset_value'},
            'test_pipeline': {'param': 'pipeline_value'},
        }

        # fallback to dataset level
        result = _get_pipeline_hyperparameter(
            hyperparameters, self.dataset_name, 'nonexistent_pipeline'
        )

        self.assertEqual(result['param'], 'dataset_value')

        # fallback to pipeline level
        result = _get_pipeline_hyperparameter(
            hyperparameters,
            'nonexistent_dataset',
            self.pipeline_name,
        )

        self.assertEqual(result['param'], 'pipeline_value')

        # fallback to default when dataset doesn't exist
        result = _get_pipeline_hyperparameter(
            hyperparameters, 'nonexistent_dataset', 'nonexistent_pipeline'
        )

        self.assertEqual(result, hyperparameters)


class TestEvaluateSignal(TestCase):
    @classmethod
    def setup_class(cls):
        cls.pipeline_name = 'test_pipeline'
        cls.signal_name = 'test_signal'
        cls.hyperparameters = {'param': 'value'}
        cls.metrics = {'f1': METRICS['f1']}
        cls.test_split = False
        cls.few_shot = False
        cls.anomaly_path = None

        cls.test_data = pd.DataFrame({'timestamp': [1, 2, 3], 'value': [1, 2, 3]})
        cls.truth_data = pd.DataFrame({'start': [1], 'end': [2]})
        cls.detected_anomalies = pd.DataFrame({'start': [1], 'end': [2], 'score': [0.9]})

    @patch('sigllm.benchmark.load_anomalies')
    @patch('sigllm.benchmark._load_signal')
    @patch('sigllm.benchmark.SigLLM')
    def test__evaluate_signal_success(self, mock_sigllm, mock_load_signal, mock_load_anomalies):
        mock_load_signal.return_value = (None, self.test_data)
        mock_load_anomalies.return_value = self.truth_data

        mock_pipeline = Mock()
        mock_pipeline.detect.return_value = self.detected_anomalies
        mock_sigllm.return_value = mock_pipeline

        result = _evaluate_signal(
            self.pipeline_name,
            self.signal_name,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            self.few_shot,
            self.anomaly_path,
        )

        assert isinstance(result, dict)
        self.assertEqual(result['status'], 'OK')
        self.assertIn('elapsed', result)
        self.assertIn('split', result)
        self.assertIn('f1', result)

        mock_load_signal.assert_called_once_with(self.signal_name, self.test_split)
        mock_load_anomalies.assert_called_once_with(self.signal_name)
        mock_sigllm.assert_called_once_with(
            self.pipeline_name, hyperparameters=self.hyperparameters
        )
        mock_pipeline.detect.assert_called_once_with(self.test_data, normal=None)

    @patch('sigllm.benchmark.load_anomalies')
    @patch('sigllm.benchmark._load_signal')
    @patch('sigllm.benchmark.SigLLM')
    def test__evaluate_signal_fail(self, mock_sigllm, mock_load_signal, mock_load_anomalies):
        mock_load_signal.return_value = (None, self.test_data)
        mock_load_anomalies.return_value = self.truth_data
        mock_sigllm.side_effect = Exception('Test error')

        result = _evaluate_signal(
            self.pipeline_name,
            self.signal_name,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            self.few_shot,
            self.anomaly_path,
        )

        assert isinstance(result, dict)
        self.assertEqual(result['status'], 'ERROR')
        self.assertIn('elapsed', result)
        self.assertIn('split', result)
        self.assertIn('f1', result)

        mock_load_signal.assert_called_once_with(self.signal_name, self.test_split)
        mock_load_anomalies.assert_called_once_with(self.signal_name)
        mock_sigllm.assert_called_once_with(
            self.pipeline_name, hyperparameters=self.hyperparameters
        )

    @patch('sigllm.benchmark.load_normal')
    @patch('sigllm.benchmark.load_anomalies')
    @patch('sigllm.benchmark._load_signal')
    @patch('sigllm.benchmark.SigLLM')
    def test__evaluate_signal_with_few_shot(
        self,
        mock_sigllm,
        mock_load_signal,
        mock_load_anomalies,
        mock_load_normal,
    ):
        mock_load_normal.return_value = self.test_data
        mock_load_signal.return_value = (None, self.test_data)
        mock_load_anomalies.return_value = self.truth_data

        mock_pipeline = Mock()
        mock_pipeline.detect.return_value = self.detected_anomalies
        mock_sigllm.return_value = mock_pipeline

        result = _evaluate_signal(
            self.pipeline_name,
            self.signal_name,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            few_shot=True,
            anomaly_path=self.anomaly_path,
        )

        assert isinstance(result, dict)
        self.assertEqual(result['status'], 'OK')

        mock_pipeline.detect.assert_called_once_with(self.test_data, normal=self.test_data)
        mock_load_normal.assert_called_once_with(self.signal_name)
        mock_load_signal.assert_called_once_with(self.signal_name, self.test_split)
        mock_load_anomalies.assert_called_once_with(self.signal_name)
        mock_sigllm.assert_called_once_with(
            self.pipeline_name, hyperparameters=self.hyperparameters
        )

    @patch('sigllm.benchmark.load_anomalies')
    @patch('sigllm.benchmark._load_signal')
    @patch('sigllm.benchmark.SigLLM')
    def test__evaluate_signal_with_anomaly_path(
        self, mock_sigllm, mock_load_signal, mock_load_anomalies
    ):
        anomaly_path = 'test_anomalies.csv'
        mock_load_signal.return_value = (None, self.test_data)
        mock_load_anomalies.return_value = self.truth_data

        mock_pipeline = Mock()
        mock_pipeline.detect.return_value = self.detected_anomalies
        mock_sigllm.return_value = mock_pipeline

        with patch.object(pd.DataFrame, 'to_csv') as mock_to_csv:
            _evaluate_signal(
                self.pipeline_name,
                self.signal_name,
                self.hyperparameters,
                self.metrics,
                self.test_split,
                self.few_shot,
                anomaly_path,
            )

            mock_to_csv.assert_called_once_with(anomaly_path, index=False)

    @patch('sigllm.benchmark.load_anomalies')
    @patch('sigllm.benchmark._load_signal')
    @patch('sigllm.benchmark.SigLLM')
    def test_evaluate_signal_error(self, mock_sigllm, mock_load_signal, mock_load_anomalies):
        mock_load_signal.side_effect = Exception('Test error')

        with self.assertRaises(Exception):
            result = _evaluate_signal(
                self.pipeline_name,
                self.signal_name,
                self.hyperparameters,
                self.metrics,
                self.test_split,
                self.few_shot,
                self.anomaly_path,
            )

            assert isinstance(result, dict)
            self.assertEqual(result['status'], 'ERROR')
            self.assertIn('elapsed', result)
            self.assertEqual(result['f1'], 0)

            mock_load_signal.assert_called_once_with(self.signal_name, self.test_split)
            mock_load_anomalies.assert_not_called()
            mock_sigllm.assert_not_called()

    @patch('sigllm.benchmark.load_anomalies')
    @patch('sigllm.benchmark._load_signal')
    @patch('sigllm.benchmark.SigLLM')
    def test__evaluate_signal_confusion_matrix(
        self, mock_sigllm, mock_load_signal, mock_load_anomalies
    ):
        mock_load_signal.return_value = (None, self.test_data)
        mock_load_anomalies.return_value = self.truth_data

        mock_pipeline = Mock()
        mock_pipeline.detect.return_value = self.detected_anomalies
        mock_sigllm.return_value = mock_pipeline

        scores = (0, 0, 0, 1)
        metrics_with_cm = {
            'f1': METRICS['f1'],
            'confusion_matrix': Mock(autospec=contextual_confusion_matrix, return_value=scores),
        }

        result = _evaluate_signal(
            self.pipeline_name,
            self.signal_name,
            self.hyperparameters,
            metrics_with_cm,
            self.test_split,
            self.few_shot,
            self.anomaly_path,
        )

        assert 'tp' in result
        assert 'fp' in result
        assert 'fn' in result
        assert 'tn' in result


class TestRunJob(TestCase):
    @classmethod
    def setup_class(cls):
        cls.pipeline = 'test_pipeline'
        cls.pipeline_name = 'test_pipeline'
        cls.dataset = 'test_dataset'
        cls.signal = 'test_signal'
        cls.hyperparameters = {'param': 'value'}
        cls.metrics = {'f1': METRICS['f1']}
        cls.test_split = False
        cls.few_shot = False
        cls.iteration = 0
        cls.run_id = 'test_run'
        cls.anomaly_path = None

        cls.temp_dir = tempfile.mkdtemp()
        cls.cache_dir = Path(cls.temp_dir) / 'cache'
        cls.anomaly_dir = Path(cls.temp_dir) / 'anomalies'

        os.makedirs(cls.cache_dir, exist_ok=True)

    @classmethod
    def teardown_class(cls):
        if os.path.exists(cls.temp_dir):
            shutil.rmtree(cls.temp_dir)

    @patch('sigllm.benchmark._evaluate_signal')
    def test_run_job_basic(self, mock_evaluate):
        mock_output = {'f1': 0.8, 'status': 'OK', 'elapsed': 1.0, 'split': False}
        mock_evaluate.return_value = mock_output

        args = (
            self.pipeline,
            self.pipeline_name,
            self.dataset,
            self.signal,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            self.few_shot,
            self.iteration,
            None,  # cache_dir
            None,  # anomaly_dir
            self.run_id,
        )

        result = _run_job(args)

        self.assertIsInstance(result, pd.DataFrame)

        row = result.iloc[0]
        self.assertEqual(row['pipeline'], self.pipeline_name)
        self.assertEqual(row['dataset'], self.dataset)
        self.assertEqual(row['signal'], self.signal)
        self.assertEqual(row['iteration'], self.iteration)
        self.assertEqual(row['run_id'], self.run_id)

        mock_evaluate.assert_called_once_with(
            self.pipeline,
            self.signal,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            self.few_shot,
            None,  # anomaly_path
        )

    @patch('sigllm.benchmark._evaluate_signal')
    def test_run_job_with_cache(self, mock_evaluate):
        mock_output = {'f1': 0.8, 'status': 'OK', 'elapsed': 1.0, 'split': False}
        mock_evaluate.return_value = mock_output

        args = (
            self.pipeline,
            self.pipeline_name,
            self.dataset,
            self.signal,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            self.few_shot,
            self.iteration,
            self.cache_dir,
            None,  # anomaly_dir
            self.run_id,
        )

        _run_job(args)

        expected_cache_file = (
            self.cache_dir / f'{self.pipeline_name}'
            f'_{self.signal}_{self.dataset}_{self.iteration}_{self.run_id}_scores.csv'
        )

        self.assertTrue(expected_cache_file.exists())

    @patch('sigllm.benchmark._evaluate_signal')
    def test_run_job_with_anomaly_dir(self, mock_evaluate):
        mock_output = {'f1': 0.8, 'status': 'OK', 'elapsed': 1.0, 'split': False}
        mock_evaluate.return_value = mock_output

        args = (
            self.pipeline,
            self.pipeline_name,
            self.dataset,
            self.signal,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            self.few_shot,
            self.iteration,
            None,  # cache_dir
            self.anomaly_dir,
            self.run_id,
        )

        _run_job(args)

        expected_anomaly_path = str(
            self.anomaly_dir / f'{self.pipeline_name}'
            f'_{self.signal}_{self.dataset}_{self.iteration}_anomalies.csv'
        )

        mock_evaluate.assert_called_once_with(
            self.pipeline,
            self.signal,
            self.hyperparameters,
            self.metrics,
            self.test_split,
            self.few_shot,
            expected_anomaly_path,
        )


class TestBenchmark(TestCase):
    @classmethod
    def setup_class(cls):
        cls.pipeline_name = 'test_pipeline'
        cls.dataset_name = 'test_dataset'
        cls.signal_name = 'test_signal'
        cls.run_id = 'test_run'

        cls.pipelines = {cls.pipeline_name: 'pipeline_path'}
        cls.datasets = {cls.dataset_name: [cls.signal_name]}
        cls.hyperparameters = {'param': 'value'}
        cls.metrics = {'f1': METRICS['f1']}

        cls.temp_dir = tempfile.mkdtemp()
        cls.cache_dir = Path(cls.temp_dir) / 'cache'
        cls.anomaly_dir = Path(cls.temp_dir) / 'anomalies'

        cls.expected_columns = [
            'pipeline',
            'rank',
            'dataset',
            'signal',
            'iteration',
            'f1',
            'status',
            'elapsed',
            'split',
            'run_id',
        ]

    @classmethod
    def teardown_class(cls):
        if os.path.exists(cls.temp_dir):
            shutil.rmtree(cls.temp_dir)

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_basic(self, mock_run_job):
        mock_output = pd.DataFrame({
            'pipeline': [self.pipeline_name],
            'dataset': [self.dataset_name],
            'signal': [self.signal_name],
            'iteration': [0],
            'f1': [0.8],
            'status': ['OK'],
            'elapsed': [1.0],
            'split': [False],
            'run_id': [self.run_id],
        })
        mock_run_job.return_value = mock_output

        result = benchmark(
            pipelines=self.pipelines,
            datasets=self.datasets,
            hyperparameters=self.hyperparameters,
            metrics=self.metrics,
        )

        self.assertIsInstance(result, pd.DataFrame)
        self.assertTrue(all(col in result.columns for col in self.expected_columns))
        mock_run_job.assert_called_once()

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_with_list_inputs(self, mock_run_job):
        mock_output = pd.DataFrame({
            'pipeline': [self.pipeline_name],
            'dataset': ['dataset'],
            'signal': [self.signal_name],
            'iteration': [0],
            'f1': [0.8],
            'status': ['OK'],
            'elapsed': [1.0],
            'split': [False],
            'run_id': [self.run_id],
        })
        mock_run_job.return_value = mock_output

        result = benchmark(
            pipelines=[self.pipeline_name],
            datasets=[self.signal_name],
            hyperparameters=[self.hyperparameters],
            metrics=list(self.metrics.values()),
        )

        self.assertIsInstance(result, pd.DataFrame)
        self.assertTrue(all(col in result.columns for col in self.expected_columns))

    def test_benchmark_with_metric_error(self):
        with self.assertRaises(ValueError):
            benchmark(
                pipelines=self.pipelines,
                datasets=self.datasets,
                hyperparameters=self.hyperparameters,
                metrics=['f1', 'nonexistent_metric'],
            )

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_with_cache(self, mock_run_job):
        os.makedirs(self.cache_dir, exist_ok=True)

        mock_output = pd.DataFrame({
            'pipeline': [self.pipeline_name],
            'dataset': [self.dataset_name],
            'signal': [self.signal_name],
            'iteration': [0],
            'f1': [0.8],
            'status': ['OK'],
            'elapsed': [1.0],
            'split': [False],
            'run_id': [self.run_id],
        })
        mock_run_job.return_value = mock_output

        result = benchmark(
            pipelines=self.pipelines, datasets=self.datasets, cache_dir=self.cache_dir
        )

        self.assertIsInstance(result, pd.DataFrame)
        self.assertTrue((self.cache_dir).exists())

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_with_anomaly_dir(self, mock_run_job):
        os.makedirs(self.anomaly_dir, exist_ok=True)

        mock_output = pd.DataFrame({
            'pipeline': [self.pipeline_name],
            'dataset': [self.dataset_name],
            'signal': [self.signal_name],
            'iteration': [0],
            'f1': [0.8],
            'status': ['OK'],
            'elapsed': [1.0],
            'split': [False],
            'run_id': [self.run_id],
        })
        mock_run_job.return_value = mock_output

        result = benchmark(
            pipelines=self.pipelines, datasets=self.datasets, anomaly_dir=self.anomaly_dir
        )

        self.assertIsInstance(result, pd.DataFrame)
        self.assertTrue((self.anomaly_dir).exists())

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_with_output_path(self, mock_run_job):
        output_path = os.path.join(self.temp_dir, 'output.csv')

        mock_output = pd.DataFrame({
            'pipeline': [self.pipeline_name],
            'dataset': [self.dataset_name],
            'signal': [self.signal_name],
            'iteration': [0],
            'f1': [0.8],
            'status': ['OK'],
            'elapsed': [1.0],
            'split': [False],
            'run_id': [self.run_id],
        })
        mock_run_job.return_value = mock_output

        result = benchmark(
            pipelines=self.pipelines,
            datasets=self.datasets,
            anomaly_dir=self.anomaly_dir,
            output_path=output_path,
        )

        self.assertIsInstance(result, pd.DataFrame)
        self.assertTrue(os.path.exists(output_path))

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_with_iterations(self, mock_run_job):
        mock_output = pd.DataFrame({
            'pipeline': [self.pipeline_name],
            'dataset': [self.dataset_name],
            'signal': [self.signal_name],
            'iteration': [0],
            'f1': [0.8],
            'status': ['OK'],
            'elapsed': [1.0],
            'split': [False],
            'run_id': [self.run_id],
        })
        mock_run_job.return_value = mock_output

        result = benchmark(pipelines=self.pipelines, datasets=self.datasets, iterations=3)

        self.assertIsInstance(result, pd.DataFrame)
        self.assertEqual(mock_run_job.call_count, 3)

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_empty_results(self, mock_run_job):
        empty_pipelines = ['nonexistent_pipeline']
        empty_datasets = {'nonexistent_dataset': []}

        # Run benchmark with minimal configuration
        result = benchmark(pipelines=empty_pipelines, datasets=empty_datasets)

        mock_run_job.assert_not_called()

        self.assertIsInstance(result, pd.DataFrame)
        self.assertTrue(result.empty)

    @patch('sigllm.benchmark._run_job')
    def test_benchmark_with_resume(self, mock_run_job):
        os.makedirs(self.cache_dir, exist_ok=True)

        cached_file = (
            self.cache_dir / f'{self.pipeline_name}'
            f'_{self.signal_name}_{self.dataset_name}_0_test_scores.csv'
        )

        pd.DataFrame({
            'pipeline': [self.pipeline_name],
            'dataset': [self.dataset_name],
            'signal': [self.signal_name],
            'iteration': [0],
            'f1': [0.8],
            'status': ['OK'],
            'elapsed': [1.0],
            'split': [False],
            'run_id': [self.run_id],
        }).to_csv(cached_file, index=False)

        result = benchmark(
            pipelines=self.pipelines, datasets=self.datasets, cache_dir=self.cache_dir, resume=True
        )

        mock_run_job.assert_not_called()

        self.assertIsInstance(result, pd.DataFrame)
        self.assertTrue(result.empty)



# -*- coding: utf-8 -*-

"""Data Management module.

This module contains functions that allow downloading demo data from Amazon S3,
as well as load and work with other data stored locally.
"""

import logging
import os

import pandas as pd
from orion.data import format_csv, load_csv

LOGGER = logging.getLogger(__name__)

DATA_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
BUCKET = 'sintel-sigllm'
S3_URL = 'https://{}.s3.amazonaws.com/{}'


def download_normal(name, data_path=DATA_PATH):
    """Load the CSV with the given name from S3.

    If the CSV has never been loaded before, it will be downloaded
    from the [sintel-sigllm bucket](https://sintel-sigllm.s3.amazonaws.com) or
    the S3 bucket specified following the `s3://{bucket}/path/to/the.csv` format,
    and then cached inside the `data` folder, within the `sigllm` package
    directory, and then returned.

    Otherwise, if it has been downloaded and cached before, it will be directly
    loaded from the `sigllm/data` folder without contacting S3.

    Args:
        name (str):
            Name of the CSV to load.
        data_path (str):
            Path to store data.

    Returns:
        pandas.DataFrame:
            A pandas.DataFrame is returned containing all the data.

    Raises:
        FileNotFoundError: If the normal file doesn't exist locally and can't
        be downloaded from S3.
    """
    try:
        url = None
        if name.startswith('s3://'):
            parts = name[5:].split('/', 1)
            bucket = parts[0]
            path = parts[1]
            url = S3_URL.format(bucket, path)
            filename = os.path.join(data_path, path.split('/')[-1])
        else:
            filename = os.path.join(data_path, name + '_normal.csv')
            data_path = os.path.join(data_path, os.path.dirname(name))

        if os.path.exists(filename):
            data = pd.read_csv(filename)
            return data

        url = url or S3_URL.format(BUCKET, '{}_normal.csv'.format(name))
        LOGGER.info('Downloading CSV %s from %s', name, url)

        try:
            data = pd.read_csv(url)
            os.makedirs(data_path, exist_ok=True)
            data.to_csv(filename, index=False)
            return data
        except Exception:
            error_msg = (
                f'Could not download or find normal file for {name}. '
                f'Please ensure the file exists at {filename} or can be '
                f'downloaded from {url}'
            )
            LOGGER.error(error_msg)
            raise FileNotFoundError(error_msg)

    except Exception as e:
        error_msg = f'Error processing normal file for {name}: {str(e)}'
        LOGGER.error(error_msg)
        raise FileNotFoundError(error_msg)


def load_normal(name, timestamp_column=None, value_column=None, start=None, end=None):
    """Load normal data from file or download if needed.

    Args:
        name (str):
            Name or path of the normal data.
        timestamp_column (str or int):
            Column index or name for timestamp.
        value_column (str or int):
            Column index or name for values.
        start (int or timestamp):
            Optional. If specified, this will be start of the sub-sequence.
        end (int or timestamp):
            Optional. If specified, this will be end of the sub-sequence.

    Returns:
        pandas.DataFrame:
            Loaded subsequence with `timestamp` and `value` columns.
    """
    if os.path.isfile(name):
        data = load_csv(name, timestamp_column, value_column)
    else:
        data = download_normal(name)

    data = format_csv(data)

    # handle start or end is specified
    if start or end:
        if any(data.index.isin([start, end])):
            data = data.iloc[start:end]
        else:
            mask = True
            if start is not None:
                mask &= data[timestamp_column] >= start
            if end is not None:
                mask &= data[timestamp_column] <= end
            data = data[mask]

    return data



# -*- coding: utf-8 -*-

import json
import os

import tiktoken
from openai import OpenAI
from tqdm import tqdm

PROMPT_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'gpt_messages.json')

PROMPTS = json.load(open(PROMPT_PATH))

VALID_NUMBERS = list('0123456789 ')
BIAS = 30


class GPT:
    """Prompt GPT models to detect anomalies in a time series.

    Args:
        name (str):
            Model name. Default to `'gpt-3.5-turbo'`.
        sep (str):
            String to separate each element in values. Default to `','`.
        anomalous_percent (float):
                Expected percentage of time series that are anomalous. Default to `0.5`.
        temp (float):
            Sampling temperature to use, between 0 and 2. Higher values like 0.8 will
            make the output more random, while lower values like 0.2 will make it
            more focused and deterministic. Do not use with `top_p`. Default to `1`.
        top_p (float):
            Alternative to sampling with temperature, called nucleus sampling, where the
            model considers the results of the tokens with top_p probability mass.
            So 0.1 means only the tokens comprising the top 10% probability mass are
            considered. Do not use with `temp`. Default to `1`.
        logprobs (bool):
            Whether to return the log probabilities of the output tokens or not.
            Defaults to `False`.
        top_logprobs (int):
            An integer between 0 and 20 specifying the number of most likely tokens
            to return at each token position. Default to `None`.
        samples (int):
            Number of responses to generate for each input message. Default to `10`.
        seed (int):
            Beta feature by OpenAI to sample deterministically. Default to `None`.
    """

    def __init__(
        self,
        name='gpt-3.5-turbo',
        sep=',',
        anomalous_percent=0.5,
        temp=1,
        top_p=1,
        logprobs=False,
        top_logprobs=None,
        samples=10,
        seed=None,
    ):
        self.name = name
        self.sep = sep
        self.anomalous_percent = anomalous_percent
        self.temp = temp
        self.top_p = top_p
        self.logprobs = logprobs
        self.top_logprobs = top_logprobs
        self.samples = samples
        self.seed = seed

        self.tokenizer = tiktoken.encoding_for_model(self.name)

        valid_tokens = []
        for number in VALID_NUMBERS:
            token = self.tokenizer.encode(number)
            valid_tokens.extend(token)

        valid_tokens.extend(self.tokenizer.encode(self.sep))
        self.logit_bias = {token: BIAS for token in valid_tokens}

        self.client = OpenAI()

    def detect(self, X, **kwargs):
        """Use GPT to forecast a signal.

        Args:
            X (ndarray):
                Input sequences of strings containing signal values.

        Returns:
            list, list:
                * List of detected anomalous values.
                * Optionally, a list of the output tokens' log probabilities.
        """
        input_length = len(self.tokenizer.encode(X[0]))
        max_tokens = int(input_length * float(self.anomalous_percent))

        all_responses, all_probs = [], []
        for text in tqdm(X):
            message = ' '.join([PROMPTS['user_message'], text, self.sep])
            response = self.client.chat.completions.create(
                model=self.name,
                messages=[
                    {'role': 'system', 'content': PROMPTS['system_message']},
                    {'role': 'user', 'content': message},
                ],
                max_tokens=max_tokens,
                temperature=self.temp,
                logprobs=self.logprobs,
                top_logprobs=self.top_logprobs,
                n=self.samples,
                seed=self.seed,
            )
            responses = [choice.message.content for choice in response.choices]
            if self.logprobs:
                probs = [choice.logprobs for choice in response.choices]
                all_probs.append(probs)

            all_responses.append(responses)

        if self.logprobs:
            return all_responses, all_probs

        return all_responses



# -*- coding: utf-8 -*-
import unittest

import numpy as np
from pytest import fixture

from sigllm.primitives.prompting.anomalies import (
    _clean_response,
    _parse_interval_response,
    _parse_list_response,
    find_anomalies_in_windows,
    format_anomalies,
    merge_anomalous_sequences,
    parse_anomaly_response,
    val2idx,
)


@fixture
def anomaly_list_within_seq():
    return np.array(
        [
            [np.array([0, 3]), np.array([1]), np.array([1, 2])],
            [np.array([0]), np.array([1, 4]), np.array([2, 3])],
            [np.array([0, 2]), np.array([]), np.array([0, 1])],
        ],
        dtype=object,
    )


@fixture
def anomaly_list_across_seq():
    return np.array(
        [np.array([0]), np.array([1, 2]), np.array([0, 2]), np.array([1, 2]), np.array([1])],
        dtype=object,
    )


@fixture
def first_indices():
    return np.array([0, 1, 2, 3, 4])


@fixture
def window_size():
    return 3


@fixture
def step_size():
    return 1


@fixture
def idx_list():
    return np.array([32, 545, 689, 1103, 1134])


@fixture
def anomalous_val():
    return np.array(
        [[np.array([0, 3]), np.array([])], [np.array([2]), np.array([4])]], dtype=object
    )


@fixture
def windows():
    return np.array([[0, 1, 0, 3], [3, 2, 6, 2]])


@fixture
def timestamp():
    return np.array(range(1000, 13000, 10))


def test_ano_within_windows(anomaly_list_within_seq):
    expected = np.array([np.array([1]), np.array([]), np.array([0])], dtype=object)

    result = find_anomalies_in_windows(anomaly_list_within_seq)

    for r, e in zip(result, expected):
        np.testing.assert_equal(r, e)


def test_merge_anomaly_seq(anomaly_list_across_seq, first_indices, window_size, step_size):
    expected = np.array([2, 4, 5])

    result = merge_anomalous_sequences(
        anomaly_list_across_seq, first_indices, window_size, step_size
    )

    np.testing.assert_equal(result, expected)


# val2idx
def test_val2idx(anomalous_val, windows):
    expected = np.array(
        [[np.array([0, 2, 3]), np.array([])], [np.array([1, 3]), np.array([])]], dtype=object
    )
    result = val2idx(anomalous_val, windows)

    for r_list, e_list in zip(result, expected):
        for r, e in zip(r_list, e_list):
            np.testing.assert_equal(r, e)


# timestamp2interval
def test_format_anomalies(idx_list, timestamp):
    expected = [(1000, 1820, 0), (5950, 6950, 0), (7390, 8390, 0), (11530, 12840, 0)]
    result = format_anomalies(idx_list, timestamp)

    assert expected == result


def test_clean_response_no_anomalies():
    test_cases = [
        'no anomalies',
        'NO ANOMALIES',
        '  no anomalies  ',
        'There are no anomalies in this data',
        'No anomaly detected',
        '  No anomaly  ',
    ]
    for text in test_cases:
        assert _clean_response(text) == ''


def test_clean_response_with_anomalies():
    test_cases = [
        ('[1, 2, 3]', '[1, 2, 3]'),
        ('  [1, 2, 3]  ', '[1, 2, 3]'),
        ('Anomalies found at [1, 2, 3]', 'anomalies found at [1, 2, 3]'),
        ('ANOMALIES AT [1, 2, 3]', 'anomalies at [1, 2, 3]'),
    ]
    for input_text, expected in test_cases:
        assert _clean_response(input_text) == expected


def test_parse_list_response_valid_cases():
    test_cases = [
        ('[1, 2, 3]', '1,2,3'),
        ('  [1, 2, 3]  ', '1,2,3'),
        ('Anomalies found at [1, 2, 3]', '1,2,3'),
        ('[1,2,3]', '1,2,3'),
        ('[1, 2, 3, 4, 5]', '1,2,3,4,5'),
    ]
    for input_text, expected in test_cases:
        assert _parse_list_response(input_text) == expected


def test_parse_list_response_invalid_cases():
    test_cases = [
        'no anomalies',
        '[]',
        '[ ]',
        'text with [no numbers]',
        'text with [letters, and, symbols]',
        '   ',
    ]
    for text in test_cases:
        assert _parse_list_response(text) == ''


def test_parse_list_response_edge_cases():
    test_cases = [
        ('[1,2,3,]', '1,2,3'),  # trailing comma
        ('[1,,2,3]', '1,2,3'),  # double comma
        ('[1, 2, 3], [5]', '1,2,3'),  # two lists
    ]
    for input_text, expected in test_cases:
        assert _parse_list_response(input_text) == expected


def test_parse_interval_response_valid_cases():
    test_cases = [
        ('[[1, 3]]', [1, 2, 3]),
        ('  [[1, 3]]  ', [1, 2, 3]),
        ('Anomalies found at [[1, 3]]', [1, 2, 3]),
        ('[[1, 3], [5, 7]]', [1, 2, 3, 5, 6, 7]),
        ('[[1, 3], [5, 7], [8, 9]]', [1, 2, 3, 5, 6, 7, 8, 9]),
        ('[[1, 3], [4, 6],]', [1, 2, 3, 4, 5, 6]),
        ('[[1, 2], [3]]', [1, 2]),
        ('[[1,,3]]', [1, 2, 3]),
        ('[[0, 10]]', list(range(11))),
    ]
    for input_text, expected in test_cases:
        assert _parse_interval_response(input_text) == expected


def test_parse_interval_response_invalid_cases():
    test_cases = [
        '[]',
        '[[]]',
        'text with [no numbers]',
        '[[1]]',  # single number instead of pair
        '[[1, 2, 3]]',  # triple instead of pair
    ]
    for text in test_cases:
        assert _parse_interval_response(text) == []


def test_parse_interval_response_multiple_matches():
    test_cases = [
        ('Found [[1, 3]] and [[5, 7]]', [1, 2, 3, 5, 6, 7]),
        ('[[1, 2]] in first part and [[3, 4]] in second', [1, 2, 3, 4]),
        ('Multiple intervals: [[1, 3]], [[4, 6]], [[7, 9]]', [1, 2, 3, 4, 5, 6, 7, 8, 9]),
        ('[[1, 2]] and [[1, 2]] and [[1, 2]]', [1, 2, 1, 2, 1, 2]),
    ]
    for input_text, expected in test_cases:
        assert _parse_interval_response(input_text) == expected


class ParseAnomalyResponseTest(unittest.TestCase):
    def test_no_anomalies(self):
        data = [['Answer: no anomalies'], ['Answer: no anomaly'], ['no anomaly, with extra']]
        expected = [[''], [''], ['']]

        output = parse_anomaly_response(data)
        self.assertEqual(output, expected)

    def test_single_anomaly(self):
        data = [['Answer: [123]'], ['Answer: [456]', 'answer: [789]']]
        expected = [['123'], ['456', '789']]

        output = parse_anomaly_response(data)
        self.assertEqual(output, expected)

    def test_multiple_anomalies(self):
        data = [['Answer: [123, 456, 789]'], ['Answer: [111, 222, 333]']]
        expected = [['123,456,789'], ['111,222,333']]

        output = parse_anomaly_response(data)
        self.assertEqual(output, expected)

    def test_mixed_responses(self):
        data = [['Answer: no anomalies', 'Answer: [123, 456]'], ['Answer: [789]', 'no anomaly']]
        expected = [['', '123,456'], ['789', '']]

        output = parse_anomaly_response(data)
        self.assertEqual(output, expected)

    def test_different_formats(self):
        data = [
            ['Answer: [123, 456]', 'Answer: [ 789 , 101 ]'],
            ['Answer: [1,2,3]', 'Answer: [ 4 , 5 , 6 ]'],
        ]
        expected = [['123,456', '789,101'], ['1,2,3', '4,5,6']]

        output = parse_anomaly_response(data)
        self.assertEqual(output, expected)

    def test_empty_responses(self):
        data = [[''], ['Answer: no anomalies'], ['answer'], ['no anomly']]
        expected = [[''], [''], [''], ['']]

        output = parse_anomaly_response(data)
        self.assertEqual(output, expected)

    def test_invalid_format(self):
        data = [['Answer: invalid format'], ['Answer: [123, abc]']]
        expected = [[''], ['']]

        output = parse_anomaly_response(data)
        self.assertEqual(output, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `sigllm.data` module."""

from datetime import datetime
from unittest.mock import patch

import pandas as pd
import pytest

from sigllm.data import load_normal


@pytest.fixture
def sample_data():
    return pd.DataFrame({
        'timestamp': pd.date_range(start='2023-01-01', periods=10, freq='D'),
        'value': range(10),
    })


@patch('sigllm.data.download_normal')
@patch('sigllm.data.format_csv')
def test_load_normal_without_start_end(mock_format_csv, mock_download, sample_data):
    mock_format_csv.return_value = sample_data
    mock_download.return_value = sample_data

    result = load_normal('test.csv')
    mock_download.assert_called_once()
    pd.testing.assert_frame_equal(result, sample_data)


@patch('sigllm.data.download_normal')
@patch('sigllm.data.format_csv')
def test_load_normal_with_index_based_start_end(mock_format_csv, mock_download, sample_data):
    mock_format_csv.return_value = sample_data
    mock_download.return_value = sample_data

    result = load_normal('test.csv', start=2, end=5)
    expected = sample_data.iloc[2:5]
    pd.testing.assert_frame_equal(result, expected)

    result = load_normal('test.csv', start=2)
    expected = sample_data.iloc[2:]
    pd.testing.assert_frame_equal(result, expected)

    result = load_normal('test.csv', end=5)
    expected = sample_data.iloc[:5]
    pd.testing.assert_frame_equal(result, expected)


@patch('sigllm.data.download_normal')
@patch('sigllm.data.format_csv')
def test_load_normal_with_timestamp_based_start_end(mock_format_csv, mock_download, sample_data):
    mock_format_csv.return_value = sample_data
    mock_download.return_value = sample_data

    start_date = datetime(2023, 1, 3)
    end_date = datetime(2023, 1, 6)
    result = load_normal('test.csv', timestamp_column='timestamp', start=start_date, end=end_date)

    expected = sample_data[
        (sample_data['timestamp'] >= start_date) & (sample_data['timestamp'] <= end_date)
    ]
    pd.testing.assert_frame_equal(result, expected)

    result = load_normal('test.csv', timestamp_column='timestamp', start=start_date)
    expected = sample_data[sample_data['timestamp'] >= start_date]
    pd.testing.assert_frame_equal(result, expected)

    result = load_normal('test.csv', timestamp_column='timestamp', end=end_date)
    expected = sample_data[sample_data['timestamp'] <= end_date]
    pd.testing.assert_frame_equal(result, expected)



import numpy as np
from pytest import fixture

from sigllm.primitives.prompting.timeseries_preprocessing import rolling_window_sequences


@fixture
def values():
    return np.array([0.555, 2.345, 1.501, 5.903, 9.116, 3.068, 4.678])


@fixture
def window_size():
    return 3


@fixture
def step_size():
    return 1


def test_rolling_window_sequences(values, window_size, step_size):
    expected = (
        np.array([
            [0.555, 2.345, 1.501],
            [2.345, 1.501, 5.903],
            [1.501, 5.903, 9.116],
            [5.903, 9.116, 3.068],
            [9.116, 3.068, 4.678],
        ]),
        np.array([0, 1, 2, 3, 4]),
        3,
        1,
        1,
    )

    result = rolling_window_sequences(values, window_size, step_size)

    if len(result) != len(expected):
        raise AssertionError('Tuples has different length')

    for arr1, arr2 in zip(result, expected):
        np.testing.assert_equal(arr1, arr2)


def test_rolling_window_sequences_multivariate(window_size, step_size):
    values = np.array([[0.555, 2.345], [1.501, 5.903], [9.116, 3.068], [7.432, 4.532]])

    expected = (
        np.array([
            [[0.555, 2.345], [1.501, 5.903], [9.116, 3.068]],
            [[1.501, 5.903], [9.116, 3.068], [7.432, 4.532]],
        ]),
        np.array([0, 1]),
        3,
        1,
        2,
    )

    result = rolling_window_sequences(values, window_size, step_size)

    if len(result) != len(expected):
        raise AssertionError('Tuples has different length')

    for arr1, arr2 in zip(result, expected):
        np.testing.assert_equal(arr1, arr2)



# -*- coding: utf-8 -*-

import logging

import torch
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer

LOGGER = logging.getLogger(__name__)

DEFAULT_BOS_TOKEN = '<s>'
DEFAULT_EOS_TOKEN = '</s>'
DEFAULT_UNK_TOKEN = '<unk>'
DEFAULT_PAD_TOKEN = '<pad>'

VALID_NUMBERS = list('0123456789')

DEFAULT_MODEL = 'mistralai/Mistral-7B-Instruct-v0.2'


class HF:
    """Prompt Pretrained models on HuggingFace to forecast a time series.

    Args:
        name (str):
            Model name. Default to `'mistralai/Mistral-7B-Instruct-v0.2'`.
        sep (str):
            String to separate each element in values. Default to `','`.
        steps (int):
            Number of steps ahead to forecast. Default `1`.
        temp (float):
            The value used to modulate the next token probabilities. Default to `1`.
        top_p (float):
             If set to float < 1, only the smallest set of most probable tokens with
             probabilities that add up to `top_p` or higher are kept for generation.
             Default to `1`.
        raw (bool):
            Whether to return the raw output or not. Defaults to `False`.
        samples (int):
            Number of forecasts to generate for each input message. Default to `1`.
        padding (int):
            Additional padding token to forecast to reduce short horizon predictions.
            Default to `0`.
    """

    def __init__(
        self,
        name=DEFAULT_MODEL,
        sep=',',
        steps=1,
        temp=1,
        top_p=1,
        raw=False,
        samples=1,
        padding=0,
    ):
        self.name = name
        self.sep = sep
        self.steps = steps
        self.temp = temp
        self.top_p = top_p
        self.raw = raw
        self.samples = samples
        self.padding = padding

        self.tokenizer = AutoTokenizer.from_pretrained(self.name, use_fast=False)

        # special tokens
        special_tokens_dict = dict()
        if self.tokenizer.eos_token is None:
            special_tokens_dict['eos_token'] = DEFAULT_EOS_TOKEN
        if self.tokenizer.bos_token is None:
            special_tokens_dict['bos_token'] = DEFAULT_BOS_TOKEN
        if self.tokenizer.unk_token is None:
            special_tokens_dict['unk_token'] = DEFAULT_UNK_TOKEN
        if self.tokenizer.pad_token is None:
            special_tokens_dict['pad_token'] = DEFAULT_PAD_TOKEN

        self.tokenizer.add_special_tokens(special_tokens_dict)
        self.tokenizer.pad_token = self.tokenizer.eos_token  # indicate the end of the time series

        # invalid tokens
        valid_tokens = []
        for number in VALID_NUMBERS:
            token = self.tokenizer.convert_tokens_to_ids(number)
            valid_tokens.append(token)

        valid_tokens.append(self.tokenizer.convert_tokens_to_ids(self.sep))
        self.invalid_tokens = [
            [i] for i in range(len(self.tokenizer) - 1) if i not in valid_tokens
        ]

        self.model = AutoModelForCausalLM.from_pretrained(
            self.name,
            device_map='auto',
            torch_dtype=torch.float16,
        )

        self.model.eval()
        self.model.generation_config.pad_token_id = self.tokenizer.pad_token_id

    def forecast(self, X, **kwargs):
        """Use GPT to forecast a signal.

        Args:
            X (ndarray):
                Input sequences of strings containing signal values.

        Returns:
            list, list:
                * List of forecasted signal values.
                * Optionally, a list of dictionaries for raw output.
        """
        all_responses, all_probs = [], []
        for text in tqdm(X):
            tokenized_input = self.tokenizer([text], return_tensors='pt').to('cuda')

            input_length = tokenized_input['input_ids'].shape[1]
            average_length = input_length / len(text.split(','))
            max_tokens = (average_length + self.padding) * self.steps

            generate_ids = self.model.generate(
                **tokenized_input,
                do_sample=True,
                max_new_tokens=max_tokens,
                temperature=self.temp,
                top_p=self.top_p,
                bad_words_ids=self.invalid_tokens,
                renormalize_logits=True,
                num_return_sequences=self.samples,
            )

            responses = self.tokenizer.batch_decode(
                generate_ids[:, input_length:],
                skip_special_tokens=True,
                clean_up_tokenization_spaces=False,
            )

            all_responses.append(responses)
            all_probs.append(generate_ids)

        if self.raw:
            return all_responses, all_probs

        return all_responses



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re

import numpy as np


def format_as_string(X, sep=',', space=False, single=False):
    """Format X to a list of string.

    Transform an array of integers to string(s), separated by the
    indicated separator and space. Handles two cases:
    - If single=True, treats X as a single time series (window_size, 1)
    - If single=False, treats X as multiple windows (num_windows, window_size, 1)

    Args:
        sep (str):
            String to separate each element in X. Default to `','`.
        space (bool):
            Whether to add space between each digit in the result. Default to `False`.
        single (bool):
            Whether to treat X as a single time series. If True, expects (window_size, 1)
            and returns a single string. If False, expects (num_windows, window_size, 1)
            and returns a list of strings. Default to `False`.

    Returns:
        ndarray or str:
            If single=True, returns one string representation. If single=False,
            returns a list of string representations for each window.
    """

    def _as_string(x):
        text = sep.join(list(map(str, x.flatten())))
        if space:
            text = ' '.join(text)
        return text

    if single:
        # single time series (window_size, 1)
        return _as_string(X)
    else:
        # multiple windows (num_windows, window_size, 1)
        results = list(map(_as_string, X))
        return np.array(results)


def _from_string_to_integer(text, sep=',', trunc=None, errors='ignore'):
    """Convert a text sequence consisting of digits to an array of integers."""
    nospace = re.sub(r'\s+', '', text)
    rule = f'[^0-9|{sep}]'

    if errors == 'raise':
        search = re.search(rule, nospace)
        if bool(search):
            start, end = search.span()
            raise ValueError(f'Encountered a non-digit value {nospace[start:end]}.')

    values = list(filter(None, nospace.split(sep)))

    if errors == 'ignore':
        clean = list(filter(lambda x: not bool(re.search(rule, x)), values))

    elif errors == 'filter':
        clean = list(map(lambda x: re.sub(rule, '', x), values))

    elif errors == 'coerce':
        clean = list(map(lambda x: x if not bool(re.search(rule, x)) else np.nan, values))

    else:
        raise KeyError(f'Unknown errors strategy {errors}.')

    clean = np.array(clean, dtype=float)

    if trunc:
        clean = clean[:trunc]

    return clean


def format_as_integer(X, sep=',', trunc=None, errors='ignore'):
    """Format a nested list of text into an array of integers.

    Transforms a list of list of string input as 3-D array of integers,
    seperated by the indicated seperator and truncated based on `trunc`.
    Handles empty strings by returning empty arrays.

    Args:
        sep (str):
            String to separate each element in values. Default to `','`.
        trunc (int):
            How many values to keep from the ndarray. Default to `None`,
            which retains all values.
        errors (str):
            Strategy to deal with erroneous values (not digits). Default
            to `'ignore'`.
            - If 'ignore', then invalid values will be ignored in the result.
            - If 'filter', then invalid values will be filtered out of the string.
            - If 'raise', then encountering invalud values will raise an exception.
            - If 'coerce', then invalid values will be set as NaN.

    Returns:
        ndarray:
            An array of digits values. Empty arrays for empty strings.
    """
    result = list()
    for string_list in X:
        sample = list()
        if not isinstance(string_list, list):
            raise ValueError('Input is not a list of lists.')

        for text in string_list:
            if not text:  # empty string
                sample.append(np.array([], dtype=float))
            else:
                scalar = _from_string_to_integer(text, sep, trunc, errors)
                sample.append(scalar)

        result.append(sample)

    output = np.array(result, dtype=object)
    if output.ndim >= 3:
        output = output.astype(float)

    return output


class Float2Scalar:
    """Convert an array of float values to scalar.

    Transforms an array of floats to an array integers. With the
    option to rescale such that the minimum value becomes zero
    and you can keep certain decimal points.

        1.05, 2., 3.1, 4.8342, 5, 0 -> 105, 200, 310, 483, 500, 0

    Args:
        decimal (int):
            Number of decimal points to keep from the float representation. Default to `2`.
        rescale (bool):
            Whether to rescale the array such that the minimum value becomes 0. Default to `True`.
    """

    def __init__(self, decimal=2, rescale=True):
        self.decimal = decimal
        self.rescale = rescale
        self.minimum = None

    def fit(self, X):
        """Learn minimum value in fit data."""
        self.minimum = np.min(X)

    def transform(self, X):
        """Transform data."""
        if self.rescale:
            X = X - self.minimum

        sign = 1 * (X >= 0) - 1 * (X < 0)
        values = np.abs(X)

        values = sign * (values * 10**self.decimal).astype(int)

        return values, self.minimum, self.decimal


class Scalar2Float:
    """Convert an array of integer values to float.

    Transforms an array of integers to an array floats.
    Shift values by minimum and include a predetermined
    number of decimal points.

        105, 200, 310, 483, 500, 0 -> 1.05, 2., 3.1, 4.8342, 5, 0

    Args:
        minimum (float):
            Bias to shift the data. Captured from Float2Scalar.
        decimal (int):
            Number of decimal points to keep from the float representation. Default to `2`.
    """

    def transform(self, X, minimum=0, decimal=2):
        """Convert data from integer to float."""
        values = X * 10 ** (-decimal)

        return values + minimum



# -*- coding: utf-8 -*-

"""Data preprocessing module.

This module contains functions that prepare timeseries for a language model.
"""

import numpy as np


def rolling_window_sequences(X, window_size=500, step_size=100):
    """Create rolling window sequences out of time series data.

    This function creates an array of sequences by rolling over the input sequence.

    Args:
        X (ndarray):
            The sequence to iterate over.
        window_size (int):
            Length of window. Defaults to 500
        step_size (int):
            Indicating the number of steps to move the window forward each round. Defaults to 100

    Returns:
        ndarray, ndarray:
            * rolling window sequences.
            * first index value of each input sequence.
    """
    if X.ndim == 1:
        dim = 1
    else:
        dim = X.shape[1]

    index = range(len(X))
    out_X = list()
    X_index = list()

    start = 0
    max_start = len(X) - window_size + 1
    while start < max_start:
        end = start + window_size
        out_X.append(X[start:end])
        X_index.append(index[start])
        start = start + step_size
    return np.asarray(out_X), np.asarray(X_index), window_size, step_size, dim



# -*- coding: utf-8 -*-
import numpy as np


def outliers(predictions):
    """Remove outliers from predictions.

    Find the outliers according to Q1 and Q3 of the drawn samples. If
    the prediction value exceeds these bounds, remove it from the output.

    Args:
        predictions (ndarray):
            Predicted sequences.

    Return:
        ndarray:
            Predictions after removing outliers.
    """
    Q1, Q3 = np.percentile(predictions, [25, 75])

    IQR = Q3 - Q1
    lower_bound = Q1 - 1.5 * IQR
    upper_bound = Q3 + 1.5 * IQR

    predictions[(predictions < lower_bound) | (predictions > upper_bound)] = np.nan

    return predictions


def aggregate_rolling_window(y, step_size=1, agg='median', remove_outliers=False):
    """Aggregate a rolling window sequence.

    Convert a rolling window sequence into a flattened time series.
    Use the aggregation specified to make each timestamp a single value.

    Args:
        y (ndarray):
            Windowed sequences. Each timestamp has multiple predictions.
        step_size (int):
            Stride size used when creating the rolling windows.
        agg (string):
            String denoting the aggregation method to use. Default is "median".
        remove_outliers (bool):
            Indicator to whether remove outliers from the predictions.

    Return:
        ndarray:
            Flattened sequence.
    """
    num_windows, num_samples, pred_length = y.shape
    num_errors = pred_length + step_size * (num_windows - 1)

    if remove_outliers:
        y = outliers(y)

    method = getattr(np, agg)
    signal = []

    for i in range(num_errors):
        intermediate = []
        for j in range(max(0, i - num_errors + pred_length), min(i + 1, pred_length)):
            for k in range(num_samples):
                intermediate.append(y[i - j, k, j])

        signal.append(method(np.asarray(intermediate)))

    return np.array(signal)



# -*- coding: utf-8 -*-

import argparse
import ast
import concurrent
import gc
import json
import logging
import os
import uuid
import warnings
from copy import deepcopy
from datetime import datetime
from functools import partial
from glob import glob
from pathlib import Path

import numpy as np
import pandas as pd
import torch
import tqdm
from mlblocks import get_pipelines_paths
from orion.benchmark import _load_signal, _parse_confusion_matrix, _sort_leaderboard
from orion.data import load_anomalies
from orion.evaluation import CONTEXTUAL_METRICS as METRICS
from orion.evaluation import contextual_confusion_matrix
from orion.progress import TqdmLogger

from sigllm import SigLLM
from sigllm.data import load_normal

warnings.simplefilter('ignore')

LOGGER = logging.getLogger(__name__)

BUCKET = 'sintel-sigllm'
S3_URL = 'https://{}.s3.amazonaws.com/{}'

BENCHMARK_PATH = os.path.join(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'), 'benchmark'
)

BENCHMARK_DATA = (
    pd.read_csv(S3_URL.format(BUCKET, 'datasets.csv'), index_col=0, header=None)
    .applymap(ast.literal_eval)
    .to_dict()[1]
)
BENCHMARK_PARAMS = (
    pd.read_csv(S3_URL.format(BUCKET, 'parameters.csv'), index_col=0, header=None)
    .applymap(ast.literal_eval)
    .to_dict()[1]
)

PIPELINE_DIR = os.path.join(os.path.dirname(__file__), 'pipelines')

PIPELINES = {
    'mistral_prompter_restricted': 'mistral_prompter',
    'mistral_prompter_0shot': 'mistral_prompter_0shot',
    'mistral_prompter_1shot': 'mistral_prompter_1shot',
}


def _get_pipeline_directory(pipeline_name):
    if os.path.isfile(pipeline_name):
        return os.path.dirname(pipeline_name)

    pipelines_paths = get_pipelines_paths()
    for base_path in pipelines_paths:
        parts = pipeline_name.split('.')
        number_of_parts = len(parts)

        for folder_parts in range(number_of_parts):
            folder = os.path.join(base_path, *parts[:folder_parts])
            filename = '.'.join(parts[folder_parts:]) + '.json'
            json_path = os.path.join(folder, filename)

            if os.path.isfile(json_path):
                return os.path.dirname(json_path)


def _get_pipeline_hyperparameter(hyperparameters, dataset_name, pipeline_name):
    hyperparameters_ = deepcopy(hyperparameters)

    if isinstance(hyperparameters, dict):
        hyperparameters_ = hyperparameters_.get(dataset_name) or hyperparameters_
        hyperparameters_ = hyperparameters_.get(pipeline_name) or hyperparameters_

    elif isinstance(hyperparameters_, str) and os.path.exists(hyperparameters_):
        with open(hyperparameters_) as f:
            hyperparameters_ = json.load(f)

    elif hyperparameters_ is None and dataset_name and pipeline_name:
        pipeline_path = _get_pipeline_directory(pipeline_name)
        if pipeline_path:
            pipeline_dirname = os.path.basename(pipeline_path)
            file_path = os.path.join(
                pipeline_path, pipeline_dirname + '_' + dataset_name.lower() + '.json'
            )
            if os.path.exists(file_path):
                hyperparameters_ = json.load(file_path)

    return hyperparameters_


def _augment_hyperparameters(hyperparameters, few_shot):
    hyperparameters_ = deepcopy(hyperparameters)
    if few_shot:
        for hyperparameter, value in hyperparameters.items():
            if 'time_segments_aggregate' in hyperparameter:
                name = hyperparameter[:-1]
                number = int(hyperparameter[-1]) + 1
                hyperparameters_[name + str(number)] = value

    return hyperparameters_


def _evaluate_signal(
    pipeline, signal, hyperparameter, metrics, test_split=False, few_shot=False, anomaly_path=None
):
    _, test = _load_signal(signal, test_split)
    truth = load_anomalies(signal)

    normal = None
    if few_shot:
        normal = load_normal(signal)

    try:
        LOGGER.info(
            'Scoring pipeline %s on signal %s (test split: %s)', pipeline, signal, test_split
        )

        start = datetime.utcnow()
        pipeline = SigLLM(pipeline, hyperparameters=hyperparameter)
        anomalies = pipeline.detect(test, normal=normal)
        elapsed = datetime.utcnow() - start

        scores = {name: scorer(truth, anomalies, test) for name, scorer in metrics.items()}

        status = 'OK'

    except Exception as ex:
        LOGGER.exception(
            'Exception scoring pipeline %s on signal %s (test split: %s), error %s.',
            pipeline,
            signal,
            test_split,
            ex,
        )

        elapsed = datetime.utcnow() - start
        anomalies = pd.DataFrame([], columns=['start', 'end', 'score'])
        scores = {name: 0 for name in metrics.keys()}

        status = 'ERROR'

    if 'confusion_matrix' in metrics.keys():
        _parse_confusion_matrix(scores, truth)

    scores['status'] = status
    scores['elapsed'] = elapsed.total_seconds()
    scores['split'] = test_split

    if anomaly_path:
        anomalies.to_csv(anomaly_path, index=False)

    del pipeline
    torch.cuda.empty_cache()
    gc.collect()
    return scores


def _run_job(args):
    # Reset random seed
    np.random.seed()

    (
        pipeline,
        pipeline_name,
        dataset,
        signal,
        hyperparameter,
        metrics,
        test_split,
        few_shot,
        iteration,
        cache_dir,
        anomaly_dir,
        run_id,
    ) = args

    anomaly_path = anomaly_dir
    if anomaly_dir:
        base_path = str(anomaly_dir / f'{pipeline_name}_{signal}_{dataset}_{iteration}')
        anomaly_path = base_path + '_anomalies.csv'

    LOGGER.info(
        'Evaluating pipeline %s on signal %s dataset %s (test split: %s); iteration %s',
        pipeline_name,
        signal,
        dataset,
        test_split,
        iteration,
    )

    output = _evaluate_signal(
        pipeline, signal, hyperparameter, metrics, test_split, few_shot, anomaly_path
    )
    scores = pd.DataFrame.from_records([output], columns=output.keys())

    scores.insert(0, 'dataset', dataset)
    scores.insert(1, 'pipeline', pipeline_name)
    scores.insert(2, 'signal', signal)
    scores.insert(3, 'iteration', iteration)
    scores['run_id'] = run_id

    if cache_dir:
        base_path = str(cache_dir / f'{pipeline_name}_{signal}_{dataset}_{iteration}_{run_id}')
        scores.to_csv(base_path + '_scores.csv', index=False)

    return scores


def benchmark(
    pipelines=None,
    datasets=None,
    hyperparameters=None,
    metrics=METRICS,
    rank='f1',
    test_split=False,
    iterations=1,
    workers=1,
    show_progress=False,
    cache_dir=None,
    anomaly_dir=None,
    resume=False,
    output_path=None,
):
    """Run pipelines on the given datasets and evaluate the performance.

    The pipelines are used to analyze the given signals and later on the
    detected anomalies are scored against the known anomalies using the
    indicated metrics.

    Finally, the scores obtained with each metric are averaged accross all the signals,
    ranked by the indicated metric and returned on a ``pandas.DataFrame``.

    Args:
        pipelines (dict or list): dictionary with pipeline names as keys and their
            JSON paths as values. If a list is given, it should be of JSON paths,
            and the paths themselves will be used as names. If not give, all verified
            pipelines will be used for evaluation.
        datasets (dict or list): dictionary of dataset name as keys and list of signals as
            values. If a list is given then it will be under a generic name ``dataset``.
            If not given, all benchmark datasets will be used used.
        hyperparameters (dict or list): dictionary with pipeline names as keys
            and their hyperparameter JSON paths or dictionaries as values. If a list is
            given, it should be of corresponding order to pipelines.
        metrics (dict or list): dictionary with metric names as keys and
            scoring functions as values. If a list is given, it should be of scoring
            functions, and they ``__name__`` value will be used as the metric name.
            If not given, all the available metrics will be used.
        rank (str): Sort and rank the pipelines based on the given metric.
            If not given, rank using the first metric.
        test_split (bool or float): Whether to use the prespecified train-test split. If
            float, then it should be between 0.0 and 1.0 and represent the proportion of
            the signal to include in the test split. If not given, use ``False``.
        iterations (int):
            Number of iterations to perform over each signal and pipeline. Defaults to 1.
        workers (int):
            If ``workers`` is given as an integer value other than 0 or 1, a multiprocessing
            Pool is used to distribute the computation across the indicated number of workers.
        show_progress (bool):
            Whether to use tqdm to keep track of the progress. Defaults to ``True``.
        cache_dir (str):
            If a ``cache_dir`` is given, intermediate results are stored in the indicated directory
            as CSV files as they get computted. This allows inspecting results while the benchmark
            is still running and also recovering results in case the process does not finish
            properly. Defaults to ``None``.
        anomaly_dir (str):
            If a ``anomaly_dir`` is given, detected anomalies will get dumped in the specificed
            directory as csv files. Defaults to ``None``.
        resume (bool):
            Whether to continue running the experiments in the benchmark from the current
            progress in ``cache_dir``.
        output_path (str): Location to save the intermediatry results. If not given,
            intermediatry results will not be saved.

    Returns:
        pandas.DataFrame:
            A table containing the scores obtained with each scoring function accross
            all the signals for each pipeline.
    """
    pipelines = pipelines or PIPELINES
    datasets = datasets or BENCHMARK_DATA
    run_id = os.getenv('RUN_ID') or str(uuid.uuid4())[:10]

    if isinstance(pipelines, list):
        pipelines = {pipeline: pipeline for pipeline in pipelines}

    if isinstance(datasets, list):
        datasets = {'dataset': datasets}

    if isinstance(hyperparameters, list):
        hyperparameters = {
            pipeline: hyperparameter
            for pipeline, hyperparameter in zip(pipelines.keys(), hyperparameters)
        }

    if isinstance(metrics, list):
        metrics_ = dict()
        for metric in metrics:
            if callable(metric):
                metrics_[metric.__name__] = metric
            elif metric in METRICS:
                metrics_[metric] = METRICS[metric]
            else:
                raise ValueError('Unknown metric: {}'.format(metric))

        metrics = metrics_

    if cache_dir:
        cache_dir = Path(cache_dir)
        os.makedirs(cache_dir, exist_ok=True)

    if anomaly_dir:
        anomaly_dir = Path(anomaly_dir)
        os.makedirs(anomaly_dir, exist_ok=True)

    jobs = list()
    for dataset, signals in datasets.items():
        for pipeline_name, pipeline in pipelines.items():
            hyperparameter = _get_pipeline_hyperparameter(hyperparameters, dataset, pipeline)
            parameters = BENCHMARK_PARAMS.get(dataset)
            few_shot = True if '1shot' in pipeline.lower() else False
            hyperparameter = _augment_hyperparameters(hyperparameter, few_shot)
            if parameters is not None:
                (test_split,) = parameters.values()

            for signal in signals:
                for iteration in range(iterations):
                    if resume:
                        experiment = str(
                            cache_dir / f'{pipeline_name}_{signal}_{dataset}_{iteration}'
                        )
                        if len(glob(experiment + '*.csv')) > 0:
                            LOGGER.warning(f'skipping {experiment}')
                            continue

                    args = (
                        pipeline,
                        pipeline_name,
                        dataset,
                        signal,
                        hyperparameter,
                        metrics,
                        test_split,
                        few_shot,
                        iteration,
                        cache_dir,
                        anomaly_dir,
                        run_id,
                    )
                    jobs.append(args)

    if workers in (0, 1):
        scores = map(_run_job, jobs)
    else:
        pool = concurrent.futures.ProcessPoolExecutor(workers)
        scores = pool.map(_run_job, jobs)

    scores = tqdm.tqdm(scores, total=len(jobs), file=TqdmLogger())
    if show_progress:
        scores = tqdm.tqdm(scores, total=len(jobs))

    if scores:
        scores = pd.concat(scores)
        if output_path:
            LOGGER.info('Saving benchmark report to %s', output_path)
            scores.to_csv(output_path, index=False)

        return _sort_leaderboard(scores, rank, metrics)

    LOGGER.info('No scores to be recorded.')
    return pd.DataFrame()


def main(pipelines, datasets, resume, workers, output_path, cache_dir, anomaly_dir, **kwargs):
    """Main to call benchmark function."""
    # output path
    output_path = os.path.join(BENCHMARK_PATH, output_path)

    # metrics
    del METRICS['accuracy']
    METRICS['confusion_matrix'] = contextual_confusion_matrix
    metrics = {k: partial(fun, weighted=False) for k, fun in METRICS.items()}

    results = benchmark(
        pipelines=pipelines,
        datasets=datasets,
        metrics=metrics,
        output_path=output_path,
        workers=workers,
        resume=resume,
        cache_dir=cache_dir,
        anomaly_dir=anomaly_dir,
    )

    return results


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('-p', '--pipelines', nargs='+', type=str, default=PIPELINES)
    parser.add_argument('-d', '--datasets', nargs='+', type=str, default=BENCHMARK_DATA)
    parser.add_argument('-r', '--resume', type=bool, default=False)
    parser.add_argument('-w', '--workers', default=1)

    parser.add_argument('-o', '--output_path', type=str, default='results.csv')
    parser.add_argument('-c', '--cache_dir', type=str, default='cache')
    parser.add_argument('-ad', '--anomaly_dir', type=str, default='anomaly_dir')

    config = parser.parse_args()

    if any([dataset in BENCHMARK_DATA.keys() for dataset in config.datasets]):
        config.datasets = dict((dataset, BENCHMARK_DATA[dataset]) for dataset in config.datasets)

    results = main(**vars(config))



# -*- coding: utf-8 -*-
import json
import logging
import os
import re

import torch
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer

PROMPT_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'huggingface_messages.json')

PROMPTS = json.load(open(PROMPT_PATH))

LOGGER = logging.getLogger(__name__)

DEFAULT_BOS_TOKEN = '<s>'
DEFAULT_EOS_TOKEN = '</s>'
DEFAULT_UNK_TOKEN = '<unk>'
DEFAULT_PAD_TOKEN = '<pad>'

VALID_NUMBERS = list('0123456789')

DEFAULT_MODEL = 'mistralai/Mistral-7B-Instruct-v0.2'


def get_examples(text, k=3):
    """Extracts the content within the first k sets of parentheses in a string.

    Args:
        text (str):
            Input string.
        k (int):
            Number of examples to return.

    Returns:
        list:
            A list containing the content within the first k sets of parentheses.
    """
    matches = re.findall(r'\(([^)]*)\)', text)
    return matches[:k]


class HF:
    """Prompt Pretrained models on HuggingFace to detect anomalies in a time series.

    Args:
        name (str):
            Model name. Default to `'mistralai/Mistral-7B-Instruct-v0.2'`.
        sep (str):
            String to separate each element in values. Default to `','`.
        anomalous_percent (float):
            Expected percentage of time series that are anomalous. Default to `0.5`.
        temp (float):
            The value used to modulate the next token probabilities. Default to `1`.
        top_p (float):
            If set to float < 1, only the smallest set of most probable tokens with
            probabilities that add up to `top_p` or higher are kept for generation.
            Default to `1`.
        raw (bool):
            Whether to return the raw output or not. Defaults to `False`.
        samples (int):
            Number of responsed to generate for each input message. Default to `10`.
        padding (int):
            Additional padding token to forecast to reduce short horizon predictions.
            Default to `0`.
        restrict_tokens (bool):
            Whether to restrict tokens or not. Default to `False`.
    """

    def __init__(
        self,
        name=DEFAULT_MODEL,
        sep=',',
        anomalous_percent=0.5,
        temp=1,
        top_p=1,
        raw=False,
        samples=10,
        padding=0,
        restrict_tokens=False,
    ):
        self.name = name
        self.sep = sep
        self.anomalous_percent = anomalous_percent
        self.temp = temp
        self.top_p = top_p
        self.raw = raw
        self.samples = samples
        self.padding = padding
        self.restrict_tokens = restrict_tokens
        self.tokenizer = AutoTokenizer.from_pretrained(self.name, use_fast=False)

        # special tokens
        special_tokens_dict = dict()
        if self.tokenizer.eos_token is None:
            special_tokens_dict['eos_token'] = DEFAULT_EOS_TOKEN
        if self.tokenizer.bos_token is None:
            special_tokens_dict['bos_token'] = DEFAULT_BOS_TOKEN
        if self.tokenizer.unk_token is None:
            special_tokens_dict['unk_token'] = DEFAULT_UNK_TOKEN
        if self.tokenizer.pad_token is None:
            special_tokens_dict['pad_token'] = DEFAULT_PAD_TOKEN

        self.tokenizer.add_special_tokens(special_tokens_dict)
        self.tokenizer.pad_token = self.tokenizer.eos_token  # indicate the end of the time series

        # Only set up invalid tokens if restriction is enabled
        if self.restrict_tokens:
            valid_tokens = []
            for number in VALID_NUMBERS:
                token = self.tokenizer.convert_tokens_to_ids(number)
                valid_tokens.append(token)

            valid_tokens.append(self.tokenizer.convert_tokens_to_ids(self.sep))
            self.invalid_tokens = [
                [i] for i in range(len(self.tokenizer) - 1) if i not in valid_tokens
            ]
        else:
            self.invalid_tokens = None

        self.model = AutoModelForCausalLM.from_pretrained(
            self.name,
            device_map='auto',
            torch_dtype=torch.float16,
        )

        self.model.eval()

    def detect(self, X, dim=1, normal=None, **kwargs):
        """Use HF to detect anomalies of a signal.

        Args:
            X (ndarray):
                Input sequences of strings containing signal values.
            dim (int, optional):
                Number of dimensions of the time series. Default to 1.
            normal (str, optional):
                A normal reference sequence for one-shot prompting. If None,
                zero-shot prompting is used. Default to None.

        Returns:
            list, list:
                * List of detected anomalous values.
                * Optionally, a list of dictionaries for raw output.
        """
        input_length = len(self.tokenizer.encode(X[0].flatten().tolist()[0]))
        max_tokens = input_length * float(self.anomalous_percent)
        all_responses, all_generate_ids = [], []

        # Prepare the one-shot example if provided
        one_shot_message = ''
        if normal is not None:
            one_shot_message = PROMPTS['one_shot_prefix'] + normal + '\n\n'

        for text in tqdm(X):
            system_message = PROMPTS['system_message']
            if self.restrict_tokens:
                user_message = PROMPTS['user_message']
            else:
                user_message = PROMPTS['user_message_unrestricted']

            if dim > 1:
                examples = get_examples(text)
                examples = ' '.join([f'({ex})' for ex in examples])

                user_message = PROMPTS['user_message_multivariate'].format(dim, examples)

            # Combine messages with one-shot example if provided
            message = ' '.join([
                system_message,
                one_shot_message,
                user_message,
                text,
                '[RESPONSE]',
            ])

            input_length = len(self.tokenizer.encode(message))

            tokenized_input = self.tokenizer(message, return_tensors='pt').to('cuda')

            generate_kwargs = {
                'do_sample': True,
                'max_new_tokens': max_tokens,
                'temperature': self.temp,
                'top_p': self.top_p,
                'renormalize_logits': True,
                'num_return_sequences': self.samples,
            }

            # Only add bad_words_ids if token restriction is enabled
            if self.restrict_tokens:
                generate_kwargs['bad_words_ids'] = self.invalid_tokens

            generate_ids = self.model.generate(**tokenized_input, **generate_kwargs)

            if self.restrict_tokens:
                responses = self.tokenizer.batch_decode(
                    generate_ids[:, input_length:],
                    skip_special_tokens=True,
                    clean_up_tokenization_spaces=False,
                )
            else:  # Extract only the part after [RESPONSE]
                # Get the full generated text
                full_responses = self.tokenizer.batch_decode(
                    generate_ids,
                    skip_special_tokens=True,
                    clean_up_tokenization_spaces=False,
                )
                responses = []
                for full_response in full_responses:
                    try:
                        response = full_response.split('[RESPONSE]')[1].strip()
                        responses.append(response)
                    except IndexError:
                        responses.append('')  # If no [RESPONSE] found, return empty string

            all_responses.append(responses)
            all_generate_ids.append(generate_ids)

        if self.raw:
            return all_responses, all_generate_ids

        return all_responses
